###########################################################
## This is an R file that forms part of the project ".."
## by Genaro Sucarrat and Alvaro Escribano.
##
## Created 19 November 2012, Madrid.
##
## -- Part 1:
## .GARCHRECURSION1   #C version (ordinary garch)
## .GARCHRECURSION1Z  #C version (ordinary garch W/ZEROS)
## rsst               #standardised skew t
##
## -- Part 2: Ordinary garch
## garch.sim          #simulate
## garch.recursion1   #recursion
## garch.logl1asym    #loglikelihood w/leverage
## garch.logl1        #loglikelihood
## garch.est          #estimation
## garch.fit          #fitted values
##
## -- Part 3: Zero-adjusted garch
## garch.recursion1z   #recursion
## garch.logl1asymz    #loglikelihood w/leverage
## garch.logl1z        #loglikelihood
## garch.estz          #estimation
## garch.fitz          #fitted values

###########################################################

##---------------------------------------------------------
## 1 LOAD LIBRARIES, COMPILE C-CODE
##---------------------------------------------------------

require(inline)
require(zoo)
#require(compiler)

#garch:
mysig <- signature(n="integer", alpha0="numeric", alpha1="numeric",
  beta1="numeric", lambda1="numeric", sigma2="numeric",
  y2="numeric", Ineg="numeric")
mycode <- "
  for (int i=0; i < *n; i++) {
    sigma2[i+1] = *alpha0 + *alpha1*y2[i] + *beta1*sigma2[i] + *lambda1*Ineg[i]*y2[i];
  }
"
.GARCHRECURSION1 <- cfunction(mysig, mycode, convention=".C") #compile

#garch w/missing values:
mysig <- signature(n="integer", alpha0="numeric", alpha1="numeric",
  beta1="numeric", lambda1="numeric", sigma2="numeric",
  y2="numeric", Ineg="numeric")
mycode <- "
  for (int i=0; i < *n; i++) {
    if(y2[i]==0){
      sigma2[i+1] = *alpha0 + *alpha1*sigma2[i] + *beta1*sigma2[i] + *lambda1*Ineg[i]*sigma2[i];
    }else{
      sigma2[i+1] = *alpha0 + *alpha1*y2[i] + *beta1*sigma2[i] + *lambda1*Ineg[i]*y2[i];
    }
  }
"
.GARCHRECURSION1Z <- cfunction(mysig, mycode, convention=".C") #compile

#Generate random draws from standardised ST density
rsst <- function(n, df=10, skew=1, zero.prob=0)
{
s2 <- sqrt(df/(df - 2))
zstar <- rt(n=n,df=df)/s2

if(skew==1){
  result <- zstar
}else{
  if(!exists("beta")){
    beta <- function(a,b){ exp( lgamma(a)+lgamma(b)-lgamma(a+b) ) }
  }
  weight <- skew/(skew + 1/skew)
  z <- runif(n, -weight, 1 - weight)
  Xi <- skew^sign(z)
  Random <- -abs(zstar)/Xi * sign(z)
  m1 <- 2 * sqrt(df - 2)/(df - 1)/beta(1/2, df/2)
  mu <- m1 * (skew - 1/skew)
  sigma <- sqrt((1 - m1^2) * (skew^2 + 1/skew^2) + 2 * m1^2 - 1)
  result <- (Random - mu)/sigma
}

if(zero.prob!=0){
  Ivar <- runif(n, -zero.prob, 1-zero.prob)
  Ivar <- (Ivar > 0)
  result <- result*Ivar/sqrt(1-zero.prob)
}

return(result)

} #end rsst

##---------------------------------------------------------
## 2 ORDINARY GARCH
##---------------------------------------------------------

#################
garch.sim <- function(n, constant=0.001, arch=0.05, garch=0.8,
  asym=0, innovations=NULL, verbose=FALSE,
  backcast.values=list(z2=NULL, sigma2=NULL))
{
if(is.null(backcast.values$z2)){ z2<-1 }else{ z2<-backcast.values$z2 }
if(is.null(innovations)){ innovations<-rnorm(n) }
z2 <- c(z2, innovations^2)
Ineg <- c(0, as.numeric(innovations < 0))
if(is.null(backcast.values$sigma2)){
  sigma2 <- constant/(1-sum(arch)-sum(garch))
}else{
  sigma2 <- backcast.values$sigma2
}
sigma2 <- c(sigma2,rep(0,n))
#zterm <- (arch + asym*Ineg)*z2 + garch

#recursion:
for(i in 1:n){
  sigma2[i+1] <- constant + arch*z2[i]*sigma2[i] + garch*sigma2[i] + asym*Ineg[i]*z2[i]*sigma2[i]
#  sigma2[i+1] <- constant + zterm[i]*sigma2[i]
}

#output:
if(verbose){
  sigma2 <- sigma2[-1]
  sigma <- sqrt(sigma2)
  y <- sigma*innovations
  Ineg <- Ineg[-1]
  result <- cbind(y, sigma, sigma2, Ineg, innovations)
}else{
  sigma2 <- sigma2[-1]
  sigma <- sqrt(sigma2)
  result <- sigma*innovations
}
return(result)
} #end garch.sim

#################
garch.recursion1 <- function(pars, aux)
{
#prepare recursion:
sigma2 <- rep(0, aux$n+1)
sigma2[1] <- pars[1]/(1-pars[2]-pars[3]) #backcast sigma2 (no asym)
y2 <- c(sigma2[1], aux$y2)
Ineg <- c(0,aux$Ineg)

#recursion:
if(aux$c.code){
  tmp <- .GARCHRECURSION1(as.integer(aux$n), as.numeric(pars[1]),
    as.numeric(pars[2]), as.numeric(pars[3]), as.numeric(pars[4]),
    as.numeric(sigma2), as.numeric(y2), as.numeric(Ineg))
  sigma2 <- tmp$sigma2
}else{
  for(i in 1:aux$n){
    sigma2[i+1] <- pars[1] + pars[2]*y2[i] + pars[3]*sigma2[i] + pars[4]*Ineg[i]*y2[i]
  }
}

#output:
return(sigma2[-1])
} #end garch.recursion1

#################
garch.logl1asym <- function(pars, aux)
{
if( any(is.na(pars)) || any(pars<=aux$lower) || any(pars>=aux$upper) ){
  chk.conds <- FALSE
}else{
  chk.conds <- TRUE
}

if(chk.conds){
  sigma2 <- garch.recursion1(pars, aux)
  sigma2 <- abs(sigma2)
  sigma2 <- pmax(.Machine$double.eps, sigma2)
  logl <- -sum(log(sigma2))/2 - aux$n*log(2*pi)/2 - sum(aux$y2/sigma2)/2
}else{ logl <- aux$logl.penalty }

return(logl)
} #end garch.logl1asym

#################
garch.logl1 <- function(pars, aux)
{
if( any(is.na(pars)) || any(pars<=aux$lower) || any(pars>=aux$upper) ){
  chk.conds <- FALSE
}else{
  chk.conds <- TRUE
}

pars <- c(pars,0)
if(chk.conds){
  sigma2 <- garch.recursion1(pars, aux)
  sigma2 <- abs(sigma2)
  sigma2 <- pmax(.Machine$double.eps, sigma2)
  logl <- -sum(log(sigma2))/2 - aux$n*log(2*pi)/2 - sum(aux$y2/sigma2)/2
}else{ logl <- aux$logl.penalty }

return(logl)
} #end garch.logl1asym


################
garch.est <- function(y, arch=1, garch=1, asym=0,
  initial.values=c(0.01, 0.05, 0.8, 0.02), gr=NULL, hessian=TRUE,
  lower=-Inf, upper=+Inf,
  method=c("Nelder-Mead", "BFGS", "CG", "L-BFGS-B", "SANN", "Brent"),
  control=list(), c.code=TRUE, logl.penalty=NULL, verbose=TRUE)
{
method <- match.arg(method)
if(verbose){
  y <- as.zoo(y)
  y <- na.trim(y)
  y.index <- index(y)
  y <- coredata(y)
}
if(asym==0){
  initial.values <- initial.values[1:3]
}

#auxiliary list:
aux <- list()
aux$n <- length(y)
aux$y2 <- y^2
aux$Ineg <- as.numeric(y < 0)
aux$lower <- lower
aux$upper <- upper
aux$c.code <- c.code
if(is.null(logl.penalty)){
  if(asym==0){
    logl.penalty <- garch.logl1(initial.values, aux)
  }else{
    logl.penalty <- garch.logl1asym(initial.values, aux)
  }
}
aux$logl.penalty <- logl.penalty

if(asym==1){
  objective.f <- function(pars,aux=aux){f <- -garch.logl1asym(pars,aux); f}
}else{
  objective.f <- function(pars,aux=aux){f <- -garch.logl1(pars,aux); f}
}

est <- optim(initial.values, objective.f, gr=gr, aux=aux, hessian=hessian,
  control=control, method=method, lower=lower, upper=upper)

#out:
est$value <- -est$value
names(est)[2] <- "objective"
if(hessian) est$hessian <- -est$hessian
if(verbose){
  parnames <- c("const", "arch1", "garch1", "asym1")[1:length(est$par)]
  names(est$par) <- parnames
  if(hessian){
    colnames(est$hessian) <- parnames
    rownames(est$hessian) <- parnames
  }
}

return(est)
} #end garch.est

#################
garch.fit <- function(y, pars, c.code=TRUE, verbose=TRUE)
{
if(verbose){
  y <- as.zoo(y)
  y <- na.trim(y)
  y.index <- index(y)
  y <- coredata(y)
}

if(length(pars)==3) pars <- c(pars,0)

#auxiliary list:
aux <- list()
aux$n <- length(y)
aux$y2 <- y^2
aux$Ineg <- as.numeric(y < 0)
aux$c.code <- c.code

sigma2 <- garch.recursion1(pars, aux)

if(verbose){
  sigma <- sqrt(sigma2)
  resid.std <- y/sigma
#  result <- cbind(y, sigma, sigma2, resid.std)
  result <- zoo(cbind(y, sigma, sigma2, resid.std), order.by=y.index)
}else{
#  result <- sigma2
  result <- zoo(sigma2, order.by=y.index)
}

return(result)
} #end garch.fit

##---------------------------------------------------------
## 3 ZERO-ADJUSTED GARCH
##---------------------------------------------------------

#################
garch.recursion1z <- function(pars, aux)
{
sigma2 <- rep(0, aux$n+1)
sigma2[1] <- pars[1]/(1-pars[2]-pars[3]) #backcast sigma2 (no asym)
y2 <- c(sigma2[1], aux$y2)
Ineg <- c(0,aux$Ineg)

#recursion:
if(aux$c.code){
  tmp <- .GARCHRECURSION1Z(as.integer(aux$n), as.numeric(pars[1]),
    as.numeric(pars[2]), as.numeric(pars[3]), as.numeric(pars[4]),
    as.numeric(sigma2), as.numeric(y2), as.numeric(Ineg))
  sigma2 <- tmp$sigma2
}else{
  for(i in 1:aux$n){
    if(y2[i]==0){ y2[i] <- sigma2[i] }
    sigma2[i+1] <- pars[1] + pars[2]*y2[i] + pars[3]*sigma2[i] + pars[4]*Ineg[i]*y2[i]
  }
}

#output:
return(sigma2[-1])
} #end garch.recursion1z

#################
garch.logl1asymz <- function(pars, aux)
{
  if( any(is.na(pars)) || any(pars<=aux$lower) || any(pars>=aux$upper) ){
    chk.conds <- FALSE
  }else{
    chk.conds <- TRUE
  }

  if(chk.conds){
    sigma2 <- garch.recursion1z(pars, aux)
    sigma2 <- abs(sigma2)
    sigma2 <- pmax(.Machine$double.eps, sigma2)
    if(aux$zeros){ sigma2 <- sigma2[-aux$zeros.where.adj] }
    logl <- -sum(log(sigma2))/2 - aux$nadj.adj*log(2*pi)/2 - sum(aux$y2adj.adj/sigma2)/2
#OLD:
#    if(aux$zeros){ sigma2 <- sigma2[-aux$zeros.where] }
#    logl <- -sum(log(sigma2))/2 - aux$nadj*log(2*pi)/2 - sum(aux$y2adj/sigma2)/2
  }else{ logl <- aux$logl.penalty }

  return(logl)
} #end garch.logl1asymz

#################
garch.logl1z <- function(pars, aux)
{
  if( any(is.na(pars)) || any(pars<=aux$lower) || any(pars>=aux$upper) ){
    chk.conds <- FALSE
  }else{
    chk.conds <- TRUE
  }

  if(aux$asym == 0){ pars <- c(pars,0) }
  if(chk.conds){
    sigma2 <- garch.recursion1z(pars, aux)
    sigma2 <- abs(sigma2)
    sigma2 <- pmax(.Machine$double.eps, sigma2)
    if(aux$zeros){ sigma2 <- sigma2[-aux$zeros.where.adj] }
    logl <- -sum(log(sigma2))/2 - aux$nadj.adj*log(2*pi)/2 - sum(aux$y2adj.adj/sigma2)/2
#OLD:
#    if(aux$zeros){ sigma2 <- sigma2[-aux$zeros.where] }
#    logl <- -sum(log(sigma2))/2 - aux$nadj*log(2*pi)/2 - sum(aux$y2adj/sigma2)/2
  }else{ logl <- aux$logl.penalty }

  return(logl)
} #end garch.logl1z

################
garch.estz <- function(y, arch=1, garch=1, asym=0,
  initial.values=c(0.01, 0.05, 0.8, 0.02), gr=NULL, hessian=TRUE,
  lower=-Inf, upper=+Inf,
  method=c("Nelder-Mead", "BFGS", "CG", "L-BFGS-B", "SANN", "Brent"),
  control=list(), c.code=TRUE, logl.penalty=NULL, verbose=TRUE,
  zero.block.size=1)
{
  method <- match.arg(method)
  if(verbose){
    y <- as.zoo(y)
    y <- na.trim(y)
    y.index <- index(y)
    y <- coredata(y)
  }
  if(asym==0){
    initial.values <- initial.values[1:3]
  }

  #auxiliary list:
  aux <- list()
  aux$asym <- asym
  aux$n <- length(y)
  aux$y2 <- y^2
  aux$Ineg <- as.numeric(y < 0)
  aux$zeros.where <- which(aux$y2==0)
  aux$zeros <- length(aux$zeros.where) > 0
  aux$nadj <- aux$n - length(aux$zeros.where)
  aux$zeros.where.adj <- aux$zeros.where
  aux$nadj.adj <- aux$nadj
  if( aux$zeros && zero.block.size > 1 ){
    #modify aux$zeros.where.adj:
    for(i in 2:zero.block.size){
      aux$zeros.where.adj <- union(aux$zeros.where.adj,
        intersect(1:aux$n, aux$zeros.where+c(i-1)))
    }
    aux$zeros.where.adj <- sort(aux$zeros.where.adj)
    #modify aux$nadj.adj:
    aux$nadj.adj <- aux$n - length(aux$zeros.where.adj)
  }
  if(aux$zeros){
    aux$y2adj <- aux$y2[-aux$zeros.where]
    aux$y2adj.adj <- aux$y2[-aux$zeros.where.adj]
  }else{
    aux$y2adj <- aux$y2
    aux$y2adj.adj <- aux$y2
  }
  aux$c.code <- c.code
  aux$lower <- -Inf
  aux$upper <- +Inf
  if(is.null(logl.penalty)){
    if(asym==0){
      logl.penalty <- garch.logl1z(initial.values, aux)
    }else{
      logl.penalty <- garch.logl1asymz(initial.values, aux)
    }
  }
  aux$logl.penalty <- logl.penalty

  #estimate:
  if(asym==1){
    objective.f <- function(pars,aux=aux){f <- -garch.logl1asymz(pars,aux); f}
  }else{
    objective.f <- function(pars,aux=aux){f <- -garch.logl1z(pars,aux); f}
  }
  est <- optim(initial.values, objective.f, gr=gr, aux=aux, hessian=hessian,
    control=control, method=method, lower=lower, upper=upper)

  #out:
  est$value <- -est$value
  names(est)[2] <- "objective"
  if(hessian) est$hessian <- -est$hessian
  if(verbose){
    parnames <- c("const", "arch1", "garch1", "asym1")[1:length(est$par)]
    names(est$par) <- parnames
    if(hessian){
      colnames(est$hessian) <- parnames
      rownames(est$hessian) <- parnames
    }
  }

  return(est)
} #end garch.estz

#################
garch.fitz <- function(y, pars, c.code=TRUE, verbose=TRUE)
{
if(verbose){
  y <- as.zoo(y)
  y <- na.trim(y)
  y.index <- index(y)
  y <- coredata(y)
}

if(length(pars)==3){ pars <- c(pars,0) }

#auxiliary list:
aux <- list()
aux$n <- length(y)
aux$y2 <- y^2
aux$Ineg <- as.numeric(y < 0)
aux$c.code <- c.code

sigma2 <- garch.recursion1z(pars, aux)

if(verbose){
  sigma <- sqrt(sigma2)
  resids.std <- y/sigma
  result <- zoo(cbind(sigma, sigma2, resids.std),
    order.by=y.index)
  colnames(result) <- c("sd", "sigma2", "residuals")
}else{
  result <- zoo(sigma2, order.by=y.index)
}

return(result)
} #end garch.fitz
